-------------------------------------------------------------------------------
--=> Lists.lua
--> Manages the lists of players shown
-------------------------------------------------------------------------------
--|> NOTES
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--|> LUACHECK
-------------------------------------------------------------------------------
--> LuaCheck Settings
-- luacheck: no max line length
-- luacheck: ignore  _.* for k self
--> weizPVP Globals
-- luacheck: globals weizPVP weizPVP.Settings
--> Addon Libs Globals
-- luacheck: globals ViragDevTool_AddData
--> WoW Globals
-- luacheck: globals time table InCombatLockdown GetTime UnitLevel

-------------------------------------------------------------------------------
--|> GLOBAL TO LOCAL
-------------------------------------------------------------------------------
local weizPVP = weizPVP

-------------------------------------------------------------------------------
--|> GLOBALS
-------------------------------------------------------------------------------
weizPVP.NearbyListSize = 0
weizPVP.CurrentList = {}
weizPVP.NearbyList = {}
weizPVP.ActiveList = {}
weizPVP.InactiveList = {}
weizPVP.ActiveDeadList = {}
weizPVP.InactiveDeadList = {}

-------------------------------------------------------------------------------
--|> LOCALS
-------------------------------------------------------------------------------
local RAID_CLASS_COLORS_TABLE = {
    ["HUNTER"] = {r = 0.67, g = 0.83, b = 0.45, colorStr = "ffabd473"},
    ["WARLOCK"] = {r = 0.53, g = 0.53, b = 0.93, colorStr = "ff8788ee"},
    ["PRIEST"] = {r = 1.0, g = 1.0, b = 1.0, colorStr = "ffffffff"},
    ["PALADIN"] = {r = 0.96, g = 0.55, b = 0.73, colorStr = "fff58cba"},
    ["MAGE"] = {r = 0.25, g = 0.78, b = 0.92, colorStr = "ff3fc7eb"},
    ["ROGUE"] = {r = 1.0, g = 0.96, b = 0.41, colorStr = "fffff569"},
    ["DRUID"] = {r = 1.0, g = 0.49, b = 0.04, colorStr = "ffff7d0a"},
    ["SHAMAN"] = {r = 0.0, g = 0.44, b = 0.87, colorStr = "ff0070de"},
    ["WARRIOR"] = {r = 0.78, g = 0.61, b = 0.43, colorStr = "ffc79c6e"},
    ["DEATHKNIGHT"] = {r = 0.77, g = 0.12, b = 0.23, colorStr = "ffc41f3b"},
    ["MONK"] = {r = 0.0, g = 1.00, b = 0.59, colorStr = "ff00ff96"},
    ["DEMONHUNTER"] = {r = 0.64, g = 0.19, b = 0.79, colorStr = "ffa330c9"},
    ["UNKNOWN"] = {r = 0.2, g = 0.2, b = 0.2, colorStr = "ff222222"}
}

weizPVP.PLAYER_INACTIVE_TIMEOUT = 31 -- how long until a player is removed from the list
weizPVP.PLAYER_ACTIVE_TIMEOUT = 14 -- how long until player is marked as inactive on the list

-------------------------------------------------------------------------------
--|> FUNCTIONS
-------------------------------------------------------------------------------

--=> Manage Expirations

-------------------------------------------------------------------------------
function weizPVP:ManageExpirations()
    --> Checks Active and Inactive lists for inactivity and move them accordingly
    --> Refreshed the list if we found expirations and Updates the NearbyCount
    local expired = false
    local i
    local timestamp = GetTime()
    --> Refresh Target
    weizPVP:PlayerTargetEvent()

    if not InCombatLockdown() then --> If Not in combat
        i = 0
        for player in pairs(weizPVP.ActiveList) do
            if (timestamp - weizPVP.ActiveList[player].TimeUpdated) > weizPVP.PLAYER_ACTIVE_TIMEOUT then
                weizPVP.InactiveList[player] = weizPVP.ActiveList[player]
                weizPVP.InactiveList[player].TimeAdded = timestamp + (i * 0.001)
                weizPVP.ActiveList[player] = nil
                expired = true
                i = 1 + 1
            end
        end
        i = 0
        for player in pairs(weizPVP.ActiveDeadList) do
            if (timestamp - weizPVP.ActiveDeadList[player].TimeUpdated) > weizPVP.PLAYER_ACTIVE_TIMEOUT then
                weizPVP.InactiveDeadList[player] = weizPVP.ActiveDeadList[player]
                weizPVP.InactiveDeadList[player].TimeAdded = timestamp + (i * 0.001)
                weizPVP.ActiveDeadList[player] = nil
                expired = true
                i = 1 + 1
            end
        end
        for player in pairs(weizPVP.InactiveList) do
            if (timestamp - weizPVP.InactiveList[player].TimeUpdated) > weizPVP.PLAYER_INACTIVE_TIMEOUT then
                weizPVP.InactiveList[player] = nil
                weizPVP.NearbyList[player] = nil
                weizPVP.PlayerActiveCache[player] = nil
                weizPVP.CurrentList[player] = nil
                expired = true
            end
        end

        for player in pairs(weizPVP.InactiveDeadList) do
            if (timestamp - weizPVP.InactiveDeadList[player].TimeUpdated) > weizPVP.PLAYER_INACTIVE_TIMEOUT then
                weizPVP.InactiveDeadList[player] = nil
                weizPVP.NearbyList[player] = nil
                weizPVP.PlayerActiveCache[player] = nil
                weizPVP.CurrentList[player] = nil
                expired = true
            end
        end

        if expired then
            weizPVP:SortNearbyList()
            weizPVP:RefreshCurrentList()
            weizPVP:UpdateNearbyCount()
        end
    end
end

--=> Add Player Data

-------------------------------------------------------------------------------
function weizPVP:AddPlayerData(player, timestamp, dead)
    --> Called from an EVENT in Events.lua when we have new player data to add or update
    --> Called when we have new data to display on the list
    weizPVP:AddPlayerNearby(player, timestamp, dead)
    weizPVP:UpdateNearbyCount()
end

--=> Add Player Nearby
-------------------------------------------------------------------------------
function weizPVP:AddPlayerNearby(player, timestamp, dead)
    if not weizPVP.NearbyList[player] then --> ADDING NEW PLAYER
        weizPVP.NearbyList[player] = {}
        weizPVP.NearbyList[player].TimeUpdated = timestamp
        weizPVP.NearbyList[player].TimeAdded = timestamp
        if dead then
            weizPVP.ActiveDeadList[player] = weizPVP.ActiveDeadList[player] or {}
            weizPVP.ActiveDeadList[player].TimeUpdated = timestamp
            weizPVP.ActiveDeadList[player].TimeAdded = timestamp
        else
            weizPVP.ActiveList[player] = weizPVP.ActiveList[player] or {}
            weizPVP.ActiveList[player].TimeAdded = timestamp
            weizPVP.ActiveList[player].TimeUpdated = timestamp
        end
    elseif not weizPVP.ActiveList[player] == nil and not weizPVP.ActiveDeadList[player] then --> EXISTING PLAYER; INACTIVE
        if not InCombatLockdown() then
            if dead then
                weizPVP.ActiveDeadList[player] = weizPVP.ActiveDeadList[player] or {}
                weizPVP.ActiveDeadList[player].TimeUpdated = timestamp
                weizPVP.ActiveDeadList[player].TimeAdded = timestamp
                weizPVP.NearbyList[player].TimeUpdated = timestamp
                weizPVP.NearbyList[player].TimeAdded = timestamp
                weizPVP.ActiveList[player] = nil
            else
                weizPVP.ActiveList[player] = weizPVP.ActiveList[player] or {}
                weizPVP.ActiveList[player].TimeUpdated = timestamp
                weizPVP.ActiveList[player].TimeAdded = timestamp
                weizPVP.NearbyList[player].TimeUpdated = timestamp
                weizPVP.NearbyList[player].TimeAdded = timestamp
                weizPVP.ActiveDeadList[player] = nil
            end
            weizPVP.InactiveList[player] = nil
            weizPVP.InactiveDeadList[player] = nil
        end
    else --> EXISTING PLAYER; ACTIVE
        if dead then
            weizPVP.ActiveDeadList[player] = weizPVP.ActiveDeadList[player] or {}
            weizPVP.ActiveDeadList[player].TimeUpdated = timestamp
            weizPVP.NearbyList[player].TimeUpdated = timestamp
            weizPVP.ActiveList[player] = nil
        else
            weizPVP.ActiveList[player] = weizPVP.ActiveList[player] or {}
            weizPVP.ActiveList[player].TimeUpdated = timestamp
            weizPVP.NearbyList[player].TimeUpdated = timestamp
            weizPVP.ActiveDeadList[player] = nil
        end
        weizPVP.InactiveList[player] = nil
        weizPVP.InactiveDeadList[player] = nil
    end
    --> Sort and Refresh List
    weizPVP:SortNearbyList()
    weizPVP:RefreshCurrentList()
    weizPVP:UpdateNearbyCount()
end

--=> Refresh Current List
-------------------------------------------------------------------------------
function weizPVP:RefreshCurrentList()
    local row = 1

    for k, data in pairs(weizPVP.CurrentList) do
        if row <= weizPVP.Settings.BARS.MAXNUMBER then
            local Level = 0
            local Class = "UNKNOWN"
            local Alpha = weizPVP.Settings.BARS.ALPHA_DEFAULT
            local Guild = ""
            local Estimated = false
            local HealthPercent = 100
            local Stealth = false
            local Dead = false
            local playerData = weizPVP.PlayerActiveCache[data.player]

            --> GET PLAYER INFO
            if playerData then
                --> LEVEL
                Estimated = playerData.Estimated
                if playerData.Level ~= 0 then
                    Level = playerData.Level
                end

                --> CLASS
                if playerData.Class then
                    Class = playerData.Class
                end

                --> GUILD
                if playerData.Guild then
                    Guild = playerData.Guild
                end

                if playerData.Stealth then
                    Stealth = playerData.Stealth
                end

                --> HEALTH PERCENT
                if playerData.HealthPercent then
                    HealthPercent = playerData.HealthPercent
                end

                --> DEAD
                if playerData.Dead then
                    Dead = playerData.Dead
                end
            end

            if weizPVP.InactiveList[data.player] or weizPVP.InactiveDeadList[data.player] then
                Alpha = weizPVP.Settings.BARS.ALPHA_INACTIVE
            end
            if weizPVP.ActiveDeadList[data.player] then
                Alpha = weizPVP.Settings.BARS.ALPHA_DEAD
            end

            weizPVP:UpdateBar(row, data.player, Alpha, HealthPercent, Class, Guild, Level, Estimated, Stealth, Dead)
            row = row + 1
        end
    end

    weizPVP.ListAmountDisplayed = row - 1
    weizPVP:ManageBarsDisplayed()
end

--=> Update Bar
-------------------------------------------------------------------------------

--> Updates the StatusBar for the bar# specified
function weizPVP:UpdateBar(num, name, Alpha, HealthPercent, Class, Guild, Level, Estimated, Stealth, Dead)
    local Bar = weizPVP.Bars[num]
    local Button = weizPVP.Buttons[num]

    if not InCombatLockdown() and name then
        Button:SetAttribute("macrotext", "/target " .. name)
    end

    Button:SetAlpha(0.1)
    Bar:SetAlpha(Alpha)

    --> NAME TEXT
    local charname = string.gsub(name, "-(.*)", "")
    if string.len(name) > string.len(charname) then
        Bar.NameText:SetText(charname .. "|cFFFF00CC*|r")
    else
        Bar.NameText:SetText(name)
    end
    Bar.name = name

    if Stealth then
        if string.len(name) > string.len(charname) then
            Bar.NameText:SetText("|cFFffc863" .. charname .. "|r |cFFFF00CC*|r")
        else
            Bar.NameText:SetText("|cFFffc863" .. name .. "|r")
        end
    end

    --> LEVEL TEXT
    if Estimated and Level then --> estimated values
        if Level < weizPVP.PlayerLevel then
            if Level == 0 then
                Bar.LevelText:SetText(" |cFFFF00CC??|r") --> Level Unknown
            else
                if Level < weizPVP.PlayerLevel  - 10 then
                    Bar.LevelText:SetText(" |cFF7cd1ff" .. Level .. "|r|cFFFF00CC+|r") --> more than 10 leves under
                else
                    Bar.LevelText:SetText(" |cFF7cff7f" .. Level .. "|r|cFFFF00CC+|r") --> less than 10 leves under
                end
                if Level < weizPVP.PlayerLevel  - 20 then
                    Bar.LevelText:SetText(" |cFF7cffd1" .. Level .. "|r|cFFFF00CC+|r") --> more than 20 leves under
                end

            end
        elseif Level > weizPVP.PlayerLevel then
            Bar.LevelText:SetText(" |cFFf7694a" .. Level .. "|r|cFFFF00CC+|r") --> higher than player's level
        else
            Bar.LevelText:SetText(" |cFFffc863" .. Level .. "|r|cFFFF00CC+|r") --> is same level as player
        end
    elseif Level then --> real values, not estimates
        if Level < weizPVP.PlayerLevel then
            if Level < weizPVP.PlayerLevel  - 10 then
                Bar.LevelText:SetText(" |cFF7cd1ff" .. Level .. "|r") --> more than 10 leves under
            else
                Bar.LevelText:SetText(" |cFF7cff7f" .. Level .. "|r") --> less than 10 leves under
            end
            if Level < weizPVP.PlayerLevel  - 20 then
                Bar.LevelText:SetText(" |cFF7cffd1" .. Level .. "|r|cFFFF00CC+|r") --> more than 20 leves under
            end
        elseif Level > weizPVP.PlayerLevel then
            Bar.LevelText:SetText(" |cFFf7694a" .. Level .. "|r") --> higher than player's level
        else
            Bar.LevelText:SetText(" |cFFffc863" .. Level .. "|r") --> is same level as player
        end
    end

    --> CLASS + COLOR (BAR COLOR)
    if Class and RAID_CLASS_COLORS_TABLE[Class] then
        local classColor = RAID_CLASS_COLORS_TABLE[Class] -- get class color
        Bar:SetStatusBarColor(classColor.r, classColor.g, classColor.b, Alpha)
    end
    Bar.bg:SetVertexColor(0, 0, 0, 1)

    --> HEALTH (BAR VALUE)
    if HealthPercent then
        Bar:SetValue(HealthPercent)
    end

    --> DEAD ICON
    if Dead ~= nil and Dead == true then
        Bar.DEADIcon:Show()
    elseif Dead == false then
        Bar.DEADIcon:Hide()
    end

    --> GUILD TEXT
    -----------------------------------------------------------------
    local guildTxtLength = Bar.LevelText:GetWidth() + Bar.NameText:GetWidth() + 6
    if Bar.DEADIcon:IsShown() then
        guildTxtLength = guildTxtLength + 22
    end
    guildTxtLength = weizPVP.Bars[1]:GetWidth() - guildTxtLength
    Bar.GuildText:SetText(Guild)
    Bar.GuildText:SetWidth(guildTxtLength)
end

--=> Sort Nearby List
-------------------------------------------------------------------------------
function weizPVP:SortNearbyList()
    --> Init ActiveList
    local active = {}
    for player in pairs(weizPVP.ActiveList) do
        local position = weizPVP.NearbyList[player].TimeAdded
        if position ~= nil then
            table.insert(active, {player = player, time = position})
        end
    end

    --> Init ActiveDeadList
    local activeDead = {}
    for player in pairs(weizPVP.ActiveDeadList) do
        local position = weizPVP.NearbyList[player].TimeAdded
        if position ~= nil then
            table.insert(activeDead, {player = player, time = position})
        end
    end

    --> Init InactiveDeadList
    local inactiveDead = {}
    for player in pairs(weizPVP.InactiveDeadList) do
        local position = weizPVP.NearbyList[player].TimeAdded
        if position ~= nil then
            table.insert(inactiveDead, {player = player, time = position})
        end
    end

    --> Init InactiveList
    local inactive = {}
    for player in pairs(weizPVP.InactiveList) do
        local position = weizPVP.NearbyList[player].TimeAdded
        if position ~= nil then
            table.insert(inactive, {player = player, time = position})
        end
    end

    table.sort(
        active,
        function(a, b)
            return a.time < b.time
        end
    )

    table.sort(
        activeDead,
        function(a, b)
            return a.time < b.time
        end
    )

    table.sort(
        inactive,
        function(a, b)
            return a.time < b.time
        end
    )

    table.sort(
        inactiveDead,
        function(a, b)
            return a.time < b.time
        end
    )

    local list = {}

    for player in pairs(active) do
        table.insert(list, active[player])
    end

    for player in pairs(activeDead) do
        table.insert(list, activeDead[player])
    end

    for player in pairs(inactive) do
        table.insert(list, inactive[player])
    end
    for player in pairs(inactiveDead) do
        table.insert(list, inactiveDead[player])
    end

    weizPVP.CurrentList = list
end

--=> Manage Bars Displayed
-------------------------------------------------------------------------------
function weizPVP:ManageBarsDisplayed()
    --> Update Nearby Count if we have a new player entry
    weizPVP:UpdateNearbyCount()

    for i = 1, weizPVP.Settings.BARS.MAXNUMBER do
        if i > weizPVP.NearbyListSize then
            -- weizPVP.Bars[i]:SetAlpha(weizPVP.Bars[i]:GetAlpha())
            -- else
            weizPVP.Bars[i]:SetAlpha(0)
        end
    end
end

--=> Update Neaby Count
-------------------------------------------------------------------------------
function weizPVP:UpdateNearbyCount()
    --> Get number of entries
    local entries = 0
    -- local entriesActive = 0
    for k in pairs(weizPVP.NearbyList) do
        entries = entries + 1
    end
    weizPVP.NearbyListSize = entries

    -- for k in pairs(weizPVP.ActiveList) do
    --     entriesActive = entriesActive + 1
    -- end

    local listSize = weizPVP.NearbyListSize

    --> Upate Global
    weizPVP.NearbyCount = listSize

    if listSize > 0 then --> List Size is > 0
        if listSize > 50 then
            weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFff0000" .. listSize .. "|r") --> 50+ (RED)
        elseif listSize > 40 then
            weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFff0056" .. listSize .. "|r") --> 40+ (BRINK PINK)
        elseif listSize > 20 then
            weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFff5900" .. listSize .. "|r") --> 20+ (ORANGE)
        elseif listSize > 10 then
            weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFFFBC00" .. listSize .. "|r") --> 10+ (GOLD)
        else
            weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFFd1ff00" .. listSize .. "|r") --> 1+ (LIME GREEM)
        end
    else --> List Size is 0
        weizPVP.HeaderFrame.Text:SetText("NEARBY |cff777777:|r |cFF26FF540|r") --> 0 (GREEN)
    end
end
